using System.Diagnostics.Contracts;

namespace Matrices
{
  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// Superinterface of all matrices
  /// </summary>
  public interface IMatrix
  {
    int this[int i, int j] { get; }
    int size { get; }

    bool Equals(IMatrix other);
    IMatrix add(IMatrix m);
    IMatrix add(FullMatrix m);
    IMatrix add(DiagonalMatrix m);
  }


  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// A general matrix
  /// </summary>
  public class FullMatrix : IMatrix
  {
    private readonly int[] rep;
    public int size { get; private set; }

    public FullMatrix(int size)
    {
      this.size = size;
      rep = new int[size * size];
    }

    public FullMatrix(FullMatrix m)
      : this(m.size)
    {
      for (var i = 0; i < m.size; i++)
        for (var j = 0; j < m.size; j++)
          this[i, j] = m[i, j];
    }

    public int this[int i, int j]
    {
      get
      {
        return rep[i * size + j];
      }
      set
      {
        rep[i * size + j] = value;
      }
    }

    public bool Equals(IMatrix other)
    {
      if (other == null || other.size != size)
        return false;

      for (var i = 0; i < size; i++)
        for (var j = 0; j < size; j++)
          if (this[i, j] != other[i, j])
            return false;
      return true;
    }

    public IMatrix add(IMatrix m)
    {
      return m.add(this);
    }

    public IMatrix add(FullMatrix m)
    {
      var result = new FullMatrix(this);
      for (var i = 0; i < size; i++)
        for (var j = 0; j < size; j++)
          result[i, j] += m[i, j];
      return result;
    }

    public IMatrix add(DiagonalMatrix m)
    {
      var result = new FullMatrix(this);
      for (var i = 0; i < size; i++)
        result[i, i] += m[i, i];
      return result;
    }
  }

  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// A diagonal matrix - only the main diagonal can be non-zero
  /// </summary>
  public class DiagonalMatrix : IMatrix
  {
    private readonly int[] rep;
    public int size { get { return rep.Length; } }

    public DiagonalMatrix(int size)
    {
      rep = new int[size];
    }

    public DiagonalMatrix(DiagonalMatrix m)
      : this(m.size)
    {
      for (var i = 0; i < m.size; i++)
        rep[i] = m.rep[i];
    }

    public int this[int i, int j]
    {
      get
      {
        return (i == j) ? rep[i] : 0;
      }
      set
      {
        Contract.Requires(i == j);
        rep[i] = value;
      }
    }

    public bool Equals(IMatrix other)
    {
      if (other == null || other.size != size)
        return false;

      for (var i = 0; i < size; i++)
        for (var j = 0; j < size; j++)
          if (this[i, j] != other[i, j])
            return false;
      return true;
    }

    public IMatrix add(IMatrix m)
    {
      return m.add(this);
    }

    public IMatrix add(FullMatrix m)
    {
      return m.add(this);
    }

    public IMatrix add(DiagonalMatrix m)
    {
      var result = new DiagonalMatrix(this);
      for (var i = 0; i < size; i++)
        result[i, i] += m[i, i];
      return result;
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// An array of matrices
  /// </summary>
  public class MatrixArray
  {
    private readonly IMatrix[] matrices;
    public readonly int size;

    /// <summary>
    /// Create empty array of given size
    /// </summary>
    /// <param name="size"></param>
    public MatrixArray(int size)
    {
      this.size = size;
      matrices = new IMatrix[size];
    }

    /// <summary>
    /// Accessor mutator for specific locations
    /// </summary>
    /// <param name="i"> </param>
    public IMatrix this[int i]
    {
      get
      {
        return matrices[i];
      }
      set
      {
        matrices[i] = value;
      }
    }

    /// <summary>
    /// Add pointwise matrices from ma and return in a third array - sizes must be equal
    /// </summary>
    /// <param name="ma"></param>
    /// <returns></returns>
    public MatrixArray addAll(MatrixArray ma)
    {
      Contract.Requires(ma != null);
      Contract.Requires(size == ma.size);

      var result = new MatrixArray(size);

      for (var index = 0; index < size; index++)
      {
        Contract.Assert(this[index] != null);
        Contract.Assert(ma[index] != null);
        Contract.Assert(this[index].size == ma[index].size);
        result[index] = this[index].add(ma[index]);
      }

      return result;
    }
  }
}