using System.Diagnostics.Contracts;

namespace Matrices
{
  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// Superinterface of all matrices
  /// </summary>
  public interface IMatrix
  {
    int this[int i, int j] { get; }
    int size { get; }

    bool Equals(IMatrix other);
  }


  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// A general matrix
  /// </summary>
  public class FullMatrix : IMatrix
  {
    private readonly int[] rep;
    public int size { get; private set; }

    public FullMatrix(int size)
    {
      this.size = size;
      rep = new int[size * size];
    }

    public FullMatrix(FullMatrix m)
      : this(m.size)
    {
      for (var i = 0; i < m.size; i++)
        for (var j = 0; j < m.size; j++)
          this[i, j] = m[i, j];
    }

    public int this[int i, int j]
    {
      get
      {
        return rep[i * size + j];
      }
      set
      {
        rep[i * size + j] = value;
      }
    }

    public bool Equals(IMatrix other)
    {
      if (other == null || other.size != size)
        return false;

      for (var i = 0; i < size; i++)
        for (var j = 0; j < size; j++)
          if (this[i, j] != other[i, j])
            return false;
      return true;
    }

  }

  ///////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// A diagonal matrix - only the main diagonal can be non-zero
  /// </summary>
  public class DiagonalMatrix : IMatrix
  {
    private readonly int[] rep;
    public int size { get { return rep.Length; } }

    public DiagonalMatrix(int size)
    {
      rep = new int[size];
    }

    public DiagonalMatrix(DiagonalMatrix m)
      : this(m.size)
    {
      for (var i = 0; i < m.size; i++)
        rep[i] = m.rep[i];
    }

    public int this[int i, int j]
    {
      get
      {
        return (i == j) ? rep[i] : 0;
      }
      set
      {
        Contract.Requires(i == j);
        rep[i] = value;
      }
    }

    public bool Equals(IMatrix other)
    {
      if (other == null || other.size != size)
        return false;

      for (var i = 0; i < size; i++)
        for (var j = 0; j < size; j++)
          if (this[i, j] != other[i, j])
            return false;
      return true;
    }

  }

  ////////////////////////////////////////////////////////////////////////////////////
  /// <summary>
  /// An array of matrices
  /// </summary>
  public class MatrixArray
  {
    private readonly IMatrix[] matrices;
    public readonly int size;

    /// <summary>
    /// Create empty array of given size
    /// </summary>
    /// <param name="size"></param>
    public MatrixArray(int size)
    {
      this.size = size;
      matrices = new IMatrix[size];
    }

    /// <summary>
    /// Accessor mutator for specific locations
    /// </summary>
    /// <param name="i"> </param>
    public IMatrix this[int i]
    {
      get
      {
        return matrices[i];
      }
      set
      {
        matrices[i] = value;
      }
    }

    /// <summary>
    /// Add pointwise matrices from ma and return in a third array - sizes must be equal
    /// </summary>
    /// <param name="ma"></param>
    /// <returns></returns>
    public MatrixArray addAll(MatrixArray ma)
    {
      Contract.Requires(size == ma.size);

      var result = new MatrixArray(size);

      for (var index = 0; index < size; index++)
      {
        var m1 = this[index];
        var m2 = ma[index];
        Contract.Assert(m1.size == m2.size);

        IMatrix mi;

        var m1f = m1 as FullMatrix;
        if (m1f != null)
        {
          var m2f = m2 as FullMatrix;
          if (m2f != null)
          {
            var t = new FullMatrix(m1f);
            for (var i = 0; i < m1f.size; i++)
              for (var j = 0; j < m1f.size; j++)
                t[i, j] += m2f[i, j];
            mi = t;
          }
          else
          {
            var m2d = m2 as DiagonalMatrix;
            Contract.Assert(m2d != null);
            var t = new FullMatrix(m1f);
            for (var i = 0; i < m1f.size; i++)
              t[i, i] += m2d[i, i];
            mi = t;
          }
        } //m1 is a FullMatrix
        else
        {
          var m1d = m1 as DiagonalMatrix;
          Contract.Assert(m1d != null);
          var m2f = m2 as FullMatrix;
          if (m2f != null)
          {
            Contract.Assert(m1d.size == m2f.size);
            var t = new FullMatrix(m2f);
            for (var i = 0; i < m1d.size; i++)
              t[i, i] += m1d[i, i];
            mi = t;
          }
          else
          {
            var m2d = m2 as DiagonalMatrix;
            Contract.Assert(m2d != null);
            var t = new DiagonalMatrix(m1d);
            for (var i = 0; i < m1d.size; i++)
              t[i, i] += m2d[i, i];
            mi = t;
          }
        }
        result[index] = mi;
      }

      return result;
    }
  }
}