﻿/*
 * CREDITS: this example is based on material developed for the
 * Introduction to Programming course by Nimi Berman and Cay Horstmann
 * at San Jose State University, CA.
 * 
 * The original material can be found at:
 *    www.cs.sjsu.edu/faculty/horstman/CS46A/debugger/tutorial.html
 */

using System;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;

namespace WordAnalyzer
{
  /// <summary>
  /// A class that analyzes words.
  /// </summary>
  public class WordAnalyzer
  {
    private String _word = "";

    [ContractInvariantMethod]
    private void ObjectInvariant()
    {
      Contract.Invariant(_word != null);
    }


    /// <summary>
    /// Setter method for field word.
    /// </summary>
    /// <param name="word">the word to be analyzed</param>
    /// <exception cref="ArgumentException">Thrown when <paramref name="word"/> is equal to null.</exception>
    public void SetWord(String word)
    {
      if (word != null)
      {
        _word = word;
      }
      else
      {
        throw new ArgumentException("The word needs to be non-null.");
      }
    }

    /// <summary>
    /// Gets the first repeated character. A character is <i>repeated</i>
    /// if it occurs at least twice in adjacent positions.
    /// </summary>
    /// <example>
    /// For example, 'l' is repeated in "hollow", but 'o' is not.
    /// </example>
    /// <returns>the first repeated character, or 0 if none found</returns>
    public char FirstRepeatedCharacter()
    {
      Contract.Ensures((Contract.Result<char>() == (char) 0) ||
                       (0 <= _word.IndexOf(new string(Contract.Result<char>(), 2)) &&
                        _word.IndexOf(new string(Contract.Result<char>(), 2)) < _word.Length &&
                        Contract.ForAll(0, _word.IndexOf(new string(Contract.Result<char>(), 2)),
                                        i => _word[i] != _word[i + 1])));

      for (int i = 0; i < _word.Length; i++)
      {
        char ch = _word.ElementAt(i);
        if (i + 1 < _word.Length && ch == _word.ElementAt(i + 1))
        {
          return ch;
        }
      }
      return (char) 0;
    }

    /// <summary>
    /// Gets the first multiply occuring character. A character is <i>multiple</i>
    /// if it occurs at least twice in the word, not necessarily in adjacent positions.
    /// </summary>
    /// <example>
    /// For example, both 'o' and 'l' are multiple in "hollow", but 'h' is not.
    /// </example>
    /// <returns>the first multiple character, or 0 if none found</returns>
    public char FirstMultipleCharacter()
    {
      Contract.Ensures((Contract.Result<char>() == (char) 0) ||
                       (0 <= Find(Contract.Result<char>(), -1) &&
                        Find(Contract.Result<char>(), -1) < _word.Length &&
                        0 <= Find(Contract.Result<char>(), Find(Contract.Result<char>(), -1)) &&
                        Contract.ForAll(0, Find(Contract.Result<char>(), -1),
                                        i => Find(_word.ElementAt(i), i) == -1)));

      for (int i = 0; i < _word.Length; i++)
      {
        char ch = _word.ElementAt(i);
        if (Find(ch, i) >= 0)
        {
          return ch;
        }
      }
      return (char) 0;
    }

    [Pure]
    private int Find(char c, int pos)
    {
      Contract.Requires(-1 <= pos && pos < _word.Length);
      Contract.Ensures(_word.Substring(pos + 1).Contains(c) ==
                       (0 <= Contract.Result<int>() && _word.ElementAt(Contract.Result<int>()) == c &&
                        Contract.ForAll(pos + 1, Contract.Result<int>(), i => _word[i] != c)));

      for (int i = pos + 1; i < _word.Length; i++)
      {
        if (_word.ElementAt(i) == c)
        {
          return i;
        }
      }
      return -1;
    }

    /// <summary>
    /// Counts the groups of repeated characters.
    /// </summary>
    /// <example>
    /// For example, "mississippi!!!" has four such groups: ss, ss, pp and !!!.
    /// </example>
    /// <returns>the number of repeated character groups</returns>
    public int CountRepeatedCharacters()
    {
      Contract.Ensures(0 <= Contract.Result<int>() && Contract.Result<int>() <= _word.Length);

      int c = 0;

      for (int i = 0; i < _word.Length - 1; i++)
      {
        if (_word.ElementAt(i) == _word.ElementAt(i + 1))
        {
          // found a repetition
          if (i == 0 || _word.ElementAt(i - 1) != _word.ElementAt(i)) // it's the start
          {
            c++;
          }
        }
      }
      return c;
    }

    /// <summary>
    /// Is the word a palindrome? A palindrome is a word that is equal to its
    /// reversed representation.
    /// </summary>
    /// <example>
    /// For example, "abba" is a palindrome.
    /// </example>
    public bool IsPalindrome()
    {
      Contract.Ensures(Contract.Result<bool>() == _word.SequenceEqual(_word.Reverse()));

      for (int i = 0; i < (_word.Length - i - 1); i++)
      {
        if (_word.ElementAt(i) != _word.ElementAt(_word.Length - i - 1))
        {
          return false;
        }
      }
      return true;
    }

    /// <summary>
    /// Write the word to the <paramref name="stream"/> if it is a palindrome.
    /// </summary>
    /// <param name="stream"></param>
    public void StorePalindrome(StreamWriter stream)
    {
      Contract.Requires(stream != null);

      if (IsPalindrome())
      {
        stream.WriteLine(_word);
      }
    }

    /// <summary>
    /// Write the word to the <paramref name="path"/> if it is a palindrome.
    /// </summary>
    /// <param name="path"></param>
    public void StorePalindrome(string path)
    {
      Contract.Requires(!string.IsNullOrEmpty(path) &&
                        path.All(c => !Path.GetInvalidPathChars().Concat(" ").Contains(c)));

      if (IsPalindrome())
      {
        Contract.Assert(path.IsNormalized());
        File.WriteAllLines(path, new[] {_word});
      }
    }
  }
}